class SingletonDecorator:

  def __init__(self, klass):
    self.klass = klass
    self.instance = None

  def __call__(self, *args, **kwargs):
    if self.instance is None:
        self.instance = self.klass(*args, **kwargs)
    return self.instance



# from standard library
import uuid
import time

from collections import deque, namedtuple
from pprint      import pprint
from threading   import Thread
from queue       import PriorityQueue, Queue
from threading   import Event as ThreadEventLib
from miros.hsm   import state_method_template
from functools   import wraps

# from this package
from miros.hsm   import HsmWithQueues
from miros.event import signals, Signal
from miros.event import Event as HsmEvent
from miros.singleton import SingletonDecorator


def pp(item):
  pprint(item)


# Add to different signals to signal if they aren't there already
Signal().append("stop_fabric")
Signal().append("stop_active_object")


class SourceThreadEvent(ThreadEventLib):
  pass


ThreadEvent = SingletonDecorator(SourceThreadEvent)


class FabricEvent():
  def __init__(self, event, priority):
    self.priority = priority
    self.event    = event

  def __lt__(self, other):
    return self.priority < other.priority

  def __eq__(self, other):
    return self.priority == other.priority


class ActiveFabricSource():
  '''A (pub-sub) event dispatcher for active objects.

  To create it:
    af = ActiveFabric()  # it is a singleton

  To subscribe (do this before starting it):
    client_deque = deque(maxlen=10)
    event_a      = HsmEvent(signal=signals.A)
    af.subscribe(client_deque, event_a, queue_type='fifo')

  To publish with priority:
    event_a = HsmEvent(signal=signals.A)
    af.publish(event_a, priority=1) # where 1 has the highest priorty
                                    # by default the priorty is set to 1000

  To be greedy, and ensure your messages take priority over everything, publish
  with a very high priority and subscribe using the lifo technique:
    client_deque = deque(maxlen=10)
    event_a      = HsmEvent(signal=signals.A)

    # By subscribing with the 'lifo' technique, any publishing event that this
    # deque item cares about will be placed at the front of the queue
    af.subscribe(client_deque, event_a, queue_type='lifo')

    # By publishing with a priority 1, the supporting tasks checking the active
    # fabric will ignore all queue items of lower priority and dispatch this
    # event first.
    af.publish(event_a, priority=1)

  Typically the active fabric is started by the first active object that is
  started.  The active fabric is a singleton, so that all active objects can
  communicate with each other by communicating to it.

  The tasks managed by this class can be started and stopped (needed for
  testing)  It does this by NOT inheriting from the threading class, instead it
  uses the a more primative approach of keeping an internal object that is a thread
  which can be killed and restarted.
  '''

  def __init__(self):
    super().__init__()

    # This is used to end the tasks forever loops
    self.task_off_event     = ThreadEvent()

    # Set up two priority queues one for the fifo and one for the lifo
    self.fifo_fabric_queue  = PriorityQueue()
    self.lifo_fabric_queue  = PriorityQueue()

    # There are two different subscription registries
    self.fifo_subscriptions = {}
    self.lifo_subscriptions = {}

    # Two different threads which will pend on items in their respective
    # priority queues
    self.fifo_thread        = None
    self.lifo_thread        = None

  def start(self):
    '''start up the threads which manage the queues and subscription registry.'''

    # Get the ThreadEvent (singleton), set it so that the threads will try and
    # run forever
    self.task_off_event = ThreadEvent()
    self.task_off_event.set()

    # Initiate a thread and return it so that this object can reference it later
    def initiate_thread(thread_obj, name, thread_runner, queue, subscriptions):
      if thread_obj is None or thread_obj.is_alive() is not True:
        thread_obj = Thread(target = thread_runner,
                                     args=(self.task_off_event,
                                           queue,
                                           subscriptions))
        thread_obj.name   = name
        thread_obj.daemon = True
        thread_obj.start()
        # if we haven't done what we promised, crash the program
        assert(thread_obj.is_alive() is True)
        return thread_obj

    # Create a thread to wake up on fifo priority queue events
    self.fifo_thread = \
      initiate_thread(thread_obj     = self.fifo_thread,
                      name           = "fifo active fabric",
                      thread_runner  = self.thread_runner_fifo,
                      queue          = self.fifo_fabric_queue,
                      subscriptions  = self.fifo_subscriptions)

    # Create a thread to wake up on lifo priority queue events
    self.lifo_thread = \
      initiate_thread(thread_obj     = self.lifo_thread,
                      name           = "lifo active fabric",
                      thread_runner  = self.thread_runner_lifo,
                      queue          = self.lifo_fabric_queue,
                      subscriptions  = self.lifo_subscriptions)

  def is_alive(self):
    result = True
    if(self.fifo_thread is not None and self.lifo_thread is not None):
      result &= self.fifo_thread.is_alive()
      result &= self.lifo_thread.is_alive()
    else:
      result &= False
    return result

  def stop(self):
    '''stop the the threads in such a way that it can be restarted again'''

    # Get the ThreadEvent (singleton), clear it so that the threads will try to
    # exit once they wake up.
    task_off_event = ThreadEvent()
    task_off_event.clear()

    # Post an item to the queue to wake up the thread then join on it until it
    # completes its last run and exits
    def stop_thread(thread_obj, queue):
      if thread_obj is not None:
        if thread_obj.is_alive() is True:
          stop_fe = FabricEvent(HsmEvent(signal=signals.stop_fabric), priority=1)
          # The tasks are pending on items in their queue, wake them up with a
          # queue item, so that they can see that they need to stop running by
          # looking at their task_off_event thread-event
          queue.put(stop_fe)
          thread_obj.join()

    stop_thread(self.fifo_thread, self.fifo_fabric_queue)
    stop_thread(self.lifo_thread, self.lifo_fabric_queue)

    if self.fifo_thread is not None:
      assert(self.fifo_thread.is_alive() is False)
    if self.lifo_thread is not None:
      assert(self.lifo_thread.is_alive() is False)

  def thread_runner_fifo(self,
                     task_off_event,
                     fifo_queue,
                     fifo_subscriptions):
    '''If this was a Thread class this function would be called "run"

    This is the main execution code of the thread.  It watches to see if the
    HsmEvent() singleton has been cleared, if it has it exits its forever loop.

    '''
    fifo_item = None
    while task_off_event.is_set():

      # this locks the task, it will only run if something is in the queue
      fifo_item = fifo_queue.get()

      if fifo_item is not None:
        event = fifo_item.event
        if event.signal_name in fifo_subscriptions.keys():
          for q in fifo_subscriptions[event.signal_name]:
            # Post the inner event from your FabricEvent object.
            q.append(fifo_item.event)

      fifo_item = None
      fifo_queue.task_done()  # so that join can work

  def thread_runner_lifo(self,
                     task_off_event,
                     lifo_queue,
                     lifo_subscriptions):
    '''If this was a Thread class this function would be called "run"

    This is the main execution code of the thread.  It watches to see if the
    HsmEvent() singleton has been cleared, if it has it exits its forever loop.

    '''
    lifo_item = None
    while task_off_event.is_set():

      # this locks the task, it will only run if something is in the queue
      lifo_item = lifo_queue.get()

      if lifo_item is not None:
        event = lifo_item.event
        if event.signal_name in lifo_subscriptions.keys():
          for q in lifo_subscriptions[event.signal_name]:
            # Post the inner event from your FabricEvent object.
            q.append(lifo_item.event)

      lifo_item = None
      lifo_queue.task_done()  # so that join can work

  def subscribe(self, queue, event_or_signal, queue_type=None):
    '''subscribe a queue to an event in a 'fifo' or 'lifo' way

    There are two different ways to subscribe to an event:
      1) first in first out - fifo priority queue
      2) last in first out  - lifo priority queue

    Both methods 1 and 2 have tasks with associated priority queues.
    The fifo/lifo action only takes place at the momement the managing task puts
    an item into the external queue used as a registry object. For method 1, the
    task puts the event on the last location in the queue.  For method 2, the
    task puts the event at the beginning location of the queue.

    Typically, you would use the same external deque object to subscribe to
    events in the two different ways provided by this method.  This is best
    explained with an example:

    Example (FIFO):
      active_object_input_queue = deque(maxlen=100) # LockingDeque works too
      event_a = HsmEvent(signal=signals.A)
      event_b = HsmEvent(signal=signals.B)

      # To show are fifo in action we first push something on our deque:
      active_object_input_queue.append(event_a)

      # subscribe to b using the fifo technique
      af = ActiveFabric()
      af.subscribe(active_object_input_queue, event_b) # default to fifo
      af.start()
      af.publish(event_b)

      # show that our old item is still there
      assert(active_object_input_queue.pop().signal_name == 'A')
      # our new item was at the end of the list
      assert(active_object_input_queue.pop().signal_name == 'B')

    Example (LIFO):
      active_object_input_queue = deque(maxlen=100) # LockingDeque works too
      event_a = HsmEvent(signal=signals.A)
      event_b = HsmEvent(signal=signals.B)

      # To show are lifo in action push something onto our deque:
      active_object_input_queue.append(event_a)

      # subscribe to b using the lifo technique
      af = ActiveFabric()
      af.subscribe(active_object_input_queue, event_b, queue_type='lifo')
      af.start()
      af.publish(event_b)

      # show that the 'lifo' technique barged our event into the front of the list
      assert(active_object_input_queue.pop().signal_name == 'B')
      # our old event is still there
      assert(active_object_input_queue.pop().signal_name == 'A')

    '''
    def _subscribe(internal_queue, signal):
      if type(signal) == HsmEvent:
        signal_name = signal.signal_name
      elif type(signal) == int:
        signal_name = signals.name_for_signal(signal)

      if signal_name in internal_queue:
        registry = internal_queue[signal_name]
        # make a list of queue ids in queue_ids
        queue_ids = map(lambda x: id(x), registry)
        if id(queue) in queue_ids:
          q_index           = registry.index(queue)
          registry[q_index] = queue
        else:
          registry.append(queue)
      else:
        internal_queue[signal_name] = [queue]
        return signal_name

    if queue_type is None:
      queue_type = 'fifo'

    if queue_type is 'lifo':
      signal_name = _subscribe(self.lifo_subscriptions, event_or_signal)
    else:
      signal_name = _subscribe(self.fifo_subscriptions, event_or_signal)
    return signal_name

  def publish(self, event, priority=None):
    '''publish an event with a given priority to all subscribed queues

    Priority of 1 is the highest priority
    '''
    if priority is None:
      priority = 1000

    fefifo = FabricEvent(event, priority)
    self.lifo_fabric_queue.put(fefifo)
    felifo = FabricEvent(event, priority)
    self.fifo_fabric_queue.put(felifo)

  def clear(self):
    '''clear out all subscriptions and queues'''
    self.fifo_fabric_queue = PriorityQueue()
    self.lifo_fabric_queue = PriorityQueue()
    self.fifo_subscriptions = {}
    self.lifo_subscriptions = {}


# ActiveFabric is a singleton class
ActiveFabric = SingletonDecorator(ActiveFabricSource)


class LockingDeque():
  '''merge of some deque and Queue object features

  This provides the locking interface of the Queue and pop, popleft, append,
  appendleft and clear features of the deque.

  Example:
    import time

    ld = LockingQueue()
    def example_task(ld):
      lq.wait() # => task will stop until an item is appended
      print(lq.pop())

    thread = Thread(target=example_task,args=(ld,))
    thread.start()   # thread started and locked
    time.sleep(10)
    ld.append("bob") # thread prints "bob"
                     # thread finished
    time.sleep(0.1)
    assert(thread.is_alive() == False)

  '''

  def __init__(self, *args, **kwargs):
    self.deque         = deque(maxlen=HsmWithQueues.QUEUE_SIZE)
    self.locking_queue = Queue(maxsize=HsmWithQueues.QUEUE_SIZE)

  def get(self, block=True, timeout=None):
    '''block on the locking queue, popleft from deque'''
    return self.locking_queue.get(block, timeout)

  def wait(self, block=True, timeout=None):
    '''wait for an append/appendleft event'''
    return self.get(block, timeout)

  def popleft(self):
    return self.deque.popleft()

  def pop(self):
    return self.deque.pop()

  def append(self, item):
    if self.locking_queue.full() is False:
      # we don't care about storing items in the locking_queue, our information
      # is in the deque, the locking_queue provides the 'get' unlocking feature
      self.locking_queue.put("ready")
      self.deque.append(item)
    else:
      self.deque.rotate(1)
      self.deque.append(item)

    if self.locking_queue.qsize() < len(self.deque):
      while self.locking_queue.qsize() != len(self.deque):
        self.locking_queue.put("ready")

  def appendleft(self, item):
    if self.locking_queue.full() is False:
      # we don't care about storing items in the locking_queue, our information
      # is in the deque, the locking_queue provides the 'get' locking feature
      self.locking_queue.put("ready")
      self.deque.appendleft(item)

    if self.locking_queue.qsize() < len(self.deque):
      while self.locking_queue.qsize() != len(self.deque):
        self.locking_queue.put("ready")

  def clear(self):
    self.deque.clear()
    try:
      while(True):
        self.locking_queue.get_nowait()
    except:
      self.locking_queue.task_done()

  def task_done(self):
    self.locking_queue.task_done()  # so that join can work

  def qsize(self):
    return self.locking_queue.qsize()

  def __len__(self):
    return len(self.deque)

  def len(self):
    return len(self.deque)


class ActiveObjectOutOfPostedEventResources(Exception):
  pass


class ActiveObject(HsmWithQueues):
  def __init__(self, name=None, instrumented=None):

    if instrumented is None:
      instrumented = True

    super().__init__(instrumented)
    self.locking_deque = LockingDeque()
    # Over-write the deque in the Hsm with Queues with the one managed by the
    # LockingDeque object. This is the 'magic' in this object.  Any time a
    # post_fifo or post_lifo method within the Hsm is touched, it unknowingly uses
    # the LockingDeque object; which will also provide the 'get' method feature.
    # This 'get' method will allow our active object task to sleep until there
    # is something to do.  To understand this object, you need to grok this.
    self.queue = self.locking_deque

    # The active fabric is a singletlon that dispatches messages between all
    # active objects.  It provides a publish/subscribe infrasture, and it will
    # post directly into our locking_deque object, and as a result, provide the
    # 'get' method of this object to unlock our task.
    self.fabric = ActiveFabric()
    self.thread = None
    self.name   = name

    # the QUEUE_SIZE is defined in HsmWithQueues
    self.posted_events_queue = deque(maxlen = self.__class__.QUEUE_SIZE)
    self.PostedEventThreadSpec = namedtuple('PostedEventThreadSpec',
      [
        'event',
        'queue_type',
        'total_times',
        'deferred',
        'period',
        'task_run_event',
      ]
    )
    self.PostedEvent = namedtuple('PostedEvents',
      [
        'signal_name',
        'task_run_event',
        'uuid',
      ]
    )

  def __thread_running(self):
    if self.thread is None:
      result = False
    else:
      result = True if self.thread.is_alive() else False
    return result

  def start_thread_if_not_running(fn):
    '''start the active object thread if it is not currently running'''
    def _start_thread_if_not_running(self, *args, **kwargs):
      if self.__thread_running() is False:
        self.__start()
      return fn(self, *args, **kwargs)
    return _start_thread_if_not_running

  def append_subscribe_to_spy(fn):
    '''instrument the full spy with our subscription request'''
    @wraps(fn)
    def _append_subscribe_to_spy(self, e_or_s, queue_type='fifo'):
      if type(e_or_s) == HsmEvent:
        signal_name = e_or_s.signal_name
      elif type(e_or_s) == int:
        signal_name = signals.name_for_signal(e_or_s)
      if self.instrumented:
        self.full.spy.append("SUBSCRIBING TO:({}, TYPE:{})".format(signal_name, queue_type))
        return fn(self, e_or_s, queue_type)
    return _append_subscribe_to_spy

  @start_thread_if_not_running
  @append_subscribe_to_spy
  def subscribe(self, event_or_signal, queue_type=None):
    if queue_type is None:
      queue_type = 'fifo'
    self.fabric.subscribe(self.queue, event_or_signal, queue_type)

  def append_publish_to_spy(fn):
    '''instrument the rtc spy with our publish event'''
    @wraps(fn)
    def _append_publish_to_spy(self, e, priority=1000):
      if self.instrumented:
        self.rtc.spy.append("PUBLISH:({}, PRIORITY:{})".format(e.signal_name, priority))
        return fn(self, e, priority)
    return _append_publish_to_spy

  @append_publish_to_spy
  @start_thread_if_not_running
  def publish(self, event, priority=None):
    '''publish an event at a given priority to the active fabric'''
    if priority is None:
      priority = 1000
    self.fabric.publish(event, priority)

  @start_thread_if_not_running
  def post_fifo(self, e, period=None, times=None, deferred=None):
    '''post an event, or events to the fifo queue

    Example of posting a single event into the fifo queue:
      ao.post_fifo(Event(signal=signals.A))

    Example create a short lived thread to post a one-shot event in one second:
      thread_id = ao.post_fifo(Event(signal=signals.A), period=1.0, time=1, deferred=True)

      # to cancel this one shot
      ao.cancel_event(thread_id)

    Example to create a heart beat of 0.7 seconds starting in 0.7 seconds:
      thread_id = ao.post_fifo(Event(signal=signals.A), period=0.7, deferred=True)

      # to kill this thread:
      ao.cancel_event(thread_id)

    '''
    if times is None:
      times = 0
    if deferred is None:
      deferred = True

    thread_id = None
    '''post to the fifo of the hsm locking deque'''
    if period is None:
      super().post_fifo(e)
    else:
      # this will make another thread to post this event into our fifo
      thread_id = self.__post_event(e, times, period, deferred, queue_type='fifo')
    return thread_id

  @start_thread_if_not_running
  def post_lifo(self, e, period=None, times=None, deferred=None):
    '''post an event, or events to the lifo queue

    Example of posting a single event into the lifo queue:
      ao.post_lifo(Event(signal=signals.A))

    Example create a short lived thread to post a one-shot event in one second:
      thread_id = ao.post_lifo(Event(signal=signals.A), period=1.0, time=1, deferred=True)

      # to cancel this one shot
      ao.cancel_event(thread_id)

    Example to create a heart beat of 0.7 seconds starting in 0.7 seconds:
      thread_id = ao.post_lifo(Event(signal=signals.A), period=0.7, deferred=True)

      # to kill this thread:
      ao.cancel_event(thread_id)

    '''
    if times is None:
      times = 0
    if deferred is None:
      deferred = True

    thread_id = None
    '''post to the lifo of the hsm locking deque'''
    if period is None:
      super().post_lifo(e)
    else:
      # this will make another thread to post this event into our lifo
      thread_id = self.__post_event(e, times, period, deferred, queue_type='lifo')
    return thread_id

  def make_unique_name_based_on_start_at_function(fn):
    '''
    If the user has not specified a name for their active object, we assign one
    based on the starting function and the first 5 characters created from a
    uuid5 using the starting state name.

    '''
    def _make_unique_name_based_on_start_at_function(self, initial_state):
      if self.name is None:
        function_name = initial_state(self, HsmEvent(signal=signals.REFLECTION_SIGNAL))
        self.name = str(uuid.uuid5(uuid.NAMESPACE_DNS, function_name))[0:5]
      fn(self, initial_state)
    return _make_unique_name_based_on_start_at_function

  @make_unique_name_based_on_start_at_function
  def start_at(self, initial_state):
    '''start the active object at a given state and begin its task'''
    super().start_at(initial_state)
    self.__start()

  def __start(self):
    '''Starts an active object thread, and the active fabric if it is not running'''
    # Get the ThreadEvent (singleton), set it so that the threads will try and
    # run forever
    task_off_event = ThreadEvent()
    task_off_event.set()

    # If the active fabric is not running, turn it on so that all of our active
    # objects can talk to each other.
    if self.fabric.is_alive() is False:
      self.fabric.start()

    def start_thread(self):
      '''Starts an active object -- called within __start

      This will start an active object and the task fabric
      '''
      thread        = Thread(target = self.run_event, args=(task_off_event, self.queue))
      thread.name   = self.name
      thread.daemon = True
      thread.start()
      return thread

    if self.__thread_running() is False:
      self.thread = start_thread(self)

    # crash if we can't start our thread
    assert(self.thread.is_alive())

  def stop(self, stop_fabric=None):
    '''Stops the active object

    Calling this method will stop all active objects.
    '''
    if stop_fabric is None:
      stop_fabric = False

    task_off_event = ThreadEvent()
    task_off_event.clear()
    self.queue.append(HsmEvent(signal=signals.stop_active_object))
    self.thread.join()
    if stop_fabric:
      self.fabric.stop()

  def run_event(self, task_off_event, queue):
    '''The active object task function

    This task method waits on the locking-deque.  If the signal is not a
    stop_active_object signal it calls the hsm next_rtc method, which will pop
    the leftmost item out of the deque part of the locking-deque and dispatch it
    into the hsm.
    '''
    while task_off_event.is_set():
      queue.wait()       # wait for an event we have subcribed to
      if self.queue.deque[0].signal_name != signals.stop_active_object:
        self.next_rtc()
      queue.task_done()  # write this so that 'join' will work

  def trace(self):
    '''Output state transition information only:

    Example:
    print(chart.trace())
      [05:23:25.314420] [<state_name>] start_at(): top->hsm_queues_graph_g1_s22
      [05:23:25.314420] [<state_name>] D->(): hsm_queues_graph_g1_s22->hsm_queues_graph_g1_s1
      [05:23:25.314420] [<state_name>] E->(): hsm_queues_graph_g1_s1->hsm_queues_graph_g1_s01
      [05:23:25.314420] [<state_name>] F->(): hsm_queues_graph_g1_s01->hsm_queues_graph_g1_s2111
      [05:23:25.314420] [<state_name>] A->(): hsm_queues_graph_g1_s2111->hsm_queues_graph_g1_s321
    '''
    strace = "\n"
    for tr in self.full.trace:
      strace += self.trace_tuple_to_formatted_string(tr)
    return strace

  def __post_event(self, e, times=None, period=None, deferred=None, queue_type=None):
    '''
    The post_event method is used to post one-shots or periodic events to the
    active object.

    It constructs a task_event and a task, then starts the task.  The task will
    run periodically posting events into either the fifo or the life of the
    active object.

    Examples:
      # Post an 'A' signal event into the lifo every 1.0 seconds, 5 times.

      # On the first time, wait one second prior to posting.  This should take
      # about 6 seconds to complete
      ao.post_event(Event(signal=signals.A),
                      time=5,
                      period=1.0,
                      deferred=True,
                      queue_type='lifo')

      # Now to cancel it, you can to cancel all events with the same
      # signal_name
      time.sleep(2.0)
      ao.cancel_events(Event(signal=signals.A))

    Example:
      # Post an event, without a time or periodic constraint
      ao.post_event(Event(signal=signals.B))  # same as ao.post_fifo(Event(signal=signals.B)

    Example of posting an event with the same signal name several times:

      # construct a thread which will post signal A, 15 times every 1 second to
      # the lifo of the active object
      post_id_1 = ao.post_event(Event(signal=signals.A),
                      time=15,
                      period=1.0,
                      deferred=True,
                      queue_type='lifo')

      # construct a thread which will post signal A, 15 times every 10 seconds to
      # the fifo of the active object
      post_id_2 = ao.post_event(Event(signal=signals.A),
                      time=15,
                      period=10.0,
                      queue_type='fifo')

      # To cancel the first event posting thread and leave the second event
      # posting thread to run:
      ao.cancel_event(uuid=uuid1)

    Example of linking a posted event to a state function handler:

        @spy_on
        def some_state_function(chart, e):
          status = return_status.UNHANDLED

          if(e.signal == signals.ENTRY_SIGNAL):
            # This will cause us to transition into the other_state_function
            # once every three seconds, starting at the next rtc event
            one_shot_uuid = chart.post_event(Event(Signal=signal.TIME_OUT,
                                                   period=3.0,
                                                   queue_type='lifo'))

            # Now we graffiti this chart with the 'one_shot_uuid' attribute so
            # that we can cancel it upon exiting the state
            chart.augment(other=one_shot_uuid, name='one_shot_uuid')
            status = return_status.HANDLED

          elif(e.signal == signals.EXIT_SIGNAL):
            chart.cancel_event(uuid=chart.one_shot_uuid)
            del(chart.one_shot_uuid)
            status = return_status.HANDLED

          elif(e.signal == signals.TIME_OUT):
            status = chart.trans(other_state_function)

          else:
            status, chart.temp.fun = return_status.SUPER, chart.top

          return status

    '''
    if deferred is None:
      deferred = True
    if queue_type is None:
      queue_type = 'fifo'
    if times is None:
      times = 1
    # if our times are set to 1 and there is no period then just post our event
    # to the fifo/lifo
    if times is 1 and period is None:
      if queue_type is 'fifo':
        self.post_fifo(e, period=None)
      else:
        self.post_lifo(e, period=None)
    else:
      # create an exit event for the task, it will be shared with the
      # cancel_event/cancel_events methods, so that the task can be stopped by
      # someone using the ActiveObject api
      task_run_event = ThreadEventLib()
      task_run_event.set()

      # set up the specification for this task
      posted_event_thread_spec = \
        self.PostedEventThreadSpec(
          event=e,
          queue_type=queue_type,
          deferred=deferred,
          period=period,
          total_times=times,
          task_run_event=task_run_event,
        )

      def post_event_thread_runner(spec, deferred, times_activated):
        # We have a Event object here that can be controlled by something
        # outside of our task.  If it is cleared, then this thread will just
        # exit and disappear from the system.
        while spec.task_run_event.is_set():
          if deferred:
            time.sleep(spec.period)
          else:
            # Pretend that we waited the first time we entered this function
            # this way we can access the time.sleep on every pass through from
            # now on.
            deferred = True

          # we might have been cancelled while we were sleeping
          if spec.task_run_event.is_set() is not True:
            break

          times_activated += 1
          if spec.queue_type is 'fifo':
            self.post_fifo(spec.event)
          else:
            self.post_lifo(spec.event)

          # If we don't want to run forever we can clear our own Event
          if spec.total_times is not 0:
            if(times_activated >= spec.total_times):
              spec.task_run_event.clear()

      thread = Thread(
        target=post_event_thread_runner,
        args=(posted_event_thread_spec,
              posted_event_thread_spec.deferred,
              0),
        daemon=True
      )
      thread.name = uuid.uuid4()
      thread.start()

      # If we have run out of spots in our queue we should issue an
      # ActiveObjectOutOfPostedEventResources since it indicates a MAJOR design
      # problem
      if(len(self.posted_events_queue) < self.__class__.QUEUE_SIZE):
        # track this thread in our posted_events deque
        self.posted_events_queue.append(
          self.PostedEvent(
            e.signal_name,
            task_run_event,
            thread.name,
          )
        )
      else:
        # Have the timer thread that we just constructed shut down (we
        # can't manage it in our posted_events deque)

        # This could easily happen if the user creates posted_event items on
        # entry and doesn't cancel them upon exiting the same state (see
        # comment in this function's docstring)
        task_run_event.clear()
        raise(ActiveObjectOutOfPostedEventResources(
          "posted_events_queue size is too small for what you have asked for"))

    return thread.name

  def cancel_event(self, uuid=None):
    '''
    This will cancel an event thread that was created using the __post_event api.
    The original call to the __post_event api would have returned the uuid needed
    to cancel it with this call.

    If there are no threads managing the uuid provided, this method will do
    nothing.

    Example:

      post_id_1 = ao.post_fifo(Event(signal=signals.A),
                      time=15,
                      period=1.0,
                      deferred=True,
                      queue_type='lifo')

      ao.cancel_event(post_id_1)


    '''
    # print("cancel uuid: {}".format(uuid))
    for i in reversed(range(len(self.posted_events_queue))):
      posted_event_task_meta_data = self.posted_events_queue[-1]
      if posted_event_task_meta_data.uuid is uuid:
        # If this thread hasn't already finished, ask it to stop
        posted_event_task_meta_data.task_run_event.clear()
        # we aren't managing this thread anymore, remove it from our list
        self.posted_events_queue.pop()
        break
      else:
        self.posted_events_queue.rotate(1)

  def cancel_events(self, e):
    '''
    This will cancel all events that have the same signal name as e, that were
    posted using the __post_event.

    This will cancel all event threads which have the same signal name as 'e',
    that were created using the __post_event api.

    If there are no threads managing the signals name within the event
    provided, this method will do nothing.

    Example:

      post_id_1 = ao.post_fifo(Event(signal=signals.A),
                      time=15,
                      period=1.0,
                      deferred=True,
                      queue_type='lifo')

      post_id_2 = ao.post_fifo(Event(signal=signals.A),
                      time=15,
                      period=1.0,
                      deferred=True,
                      queue_type='lifo')

      ao.cancel_events(Event(signal=signals.A))
    '''
    # cancel all threads which could cause this event to take place
    for i in reversed(range(len(self.posted_events_queue))):
      posted_event_task_meta_data = self.posted_events_queue[-1]
      if posted_event_task_meta_data.signal_name is e.signal_name:
        # If this thread hasn't already finished, ask it to stop
        posted_event_task_meta_data.task_run_event.clear()
        # we aren't managing this thread anymore, remove it from our list
        self.posted_events_queue.pop()
      else:
        self.posted_events_queue.rotate(1)


class Factory(ActiveObject):

  class StateMethodBlueprint():
    def __init__(self, name, ao):
      self.name = name
      self.state_method = state_method_template(name)
      self.ao = ao

    def catch(self, signal, handler):
      self.ao.register_signal_callback(self.state_method, signal, handler)
      return self

    def to_method(self):
      return self.state_method

  def __init__(self, name):
    super().__init__()
    self.name = name
    self.states = {}

  def create(self, state=None):
    '''
    This will allow the Factory to create different things and attach them to
    itself, for now it can only create states.

    '''
    self.states[state] = self.__class__.StateMethodBlueprint(name=state, ao=self)
    return self.states[state]

  def nest(self, state, parent):
    if type(state) == str:
      state_method = self.state[state].state_method
    else:
      state_method = state

    if type(parent) == str:
      parent_state_method = self.state[parent].state_method
    else:
      parent_state_method = parent
      if parent_state_method is None:
        parent_state_method = self.top

    self.register_parent(state_method, parent_state_method)
    return self

  def start_at(self, state):
    if type(state) == str:
      state_method = self.state[state].state_method
    else:
      state_method = state

    super().start_at(state_method)

  def to_code(self, state):
    if type(state) == str:
      state_method = self.states[state].state_method
    else:
      state_method = state
    return super().to_code(state_method)



# -*- coding: utf-8 -*-
"""
This module provides a hierarchical state machine event class (HsmEventProcessor), and an
instrumented hierarchical state machine class (InstrumentedHsmEventProcessor).  The
InstrumentedHsmEventProcessor is inherited from the HsmEventProcessor, and it provides two different views
into the workings of your state machine:

  * spy   -> complete record of all search, transitions and hooks
  * trace -> only provides information about state transition (no hook
             information)

To define an HsmEventProcessor, you would create a number of methods outside of this class,
then inject them into the HsmEventProcessor, by calling the 'start_at' method.

Example::


                       +------- graph_b1_s1 -----s-----+
                       |  +---- graph_b1_s2 -----t-+   |
                       |  |  +- graph_b1_s3 -+     |   |
                       |  |  |               |   +-+   |
                       |  |  |               <-b-+ <-a-+
                       |  |  +---------------+     |   |
                       |  +------------------------+   |
                       +-------------------------------+

  # To create this cart, we
  # 1) import the required items from miros:
  from miros.event import ReturnStatus, signals, Event, return_status
  from miros.hsm   import InstrumentedHsmEventProcessor, HsmTopologyException, spy_on

  # 2) create the three different states:
  @spy_on
  def graph_b1_s1(chart, e):
    status = return_status.UNHANDLED
    if(e.signal == signals.ENTRY_SIGNAL):
      status = return_status.HANDLED
    elif(e.signal == signals.EXIT_SIGNAL):
      status = return_status.HANDLED
    elif(e.signal == signals.A):
      status = chart.trans(graph_b1_s2)
    else:
      status, chart.temp.fun = return_status.SUPER, chart.top
    return status

  # define an inner state who's super state is graph_b1_s1
  @spy_on
  def graph_b1_s2(chart, e):
    status = return_status.UNHANDLED
    if(e.signal == signals.ENTRY_SIGNAL):
      status = return_status.HANDLED
    elif(e.signal == signals.EXIT_SIGNAL):
      status = return_status.HANDLED
    elif(e.signal == signals.B):
      status = chart.trans(graph_b1_s3)
    else:
      status, chart.temp.fun = return_status.SUPER, graph_b1_s1
    return status

  # define an inner state who's super state is graph_b1_s2
  @spy_on
  def graph_b1_s3(chart, e):
    status = return_status.UNHANDLED
    if(e.signal == signals.ENTRY_SIGNAL):
      status = return_status.HANDLED
    elif(e.signal == signals.EXIT_SIGNAL):
      status = return_status.HANDLED
    else:
      status, chart.temp.fun = return_status.SUPER, graph_b1_s2
    return status

  # 3) Create an HsmEventProcessor, in this case we make one that is instrumented (a bit
  # slower than a plain HsmEventProcessor, but we can use it to see what happened)
  chart = InstrumentedHsmEventProcessor()

  # 4) Start the chart in the state we desire
  chart.start_at(graph_b1_s1)

  # 5) Send an Event(s) to the chart:
  chart.dispatch(Event(signal=signals.A)

  # 6) Look at what happened
  import pprint
  def pp(item):
    pprint.pprint(item)
  pp(chart.full.spy)
                     # ['START',
                     #  'SEARCH_FOR_SUPER_SIGNAL:graph_b1_s2',
                     #  'SEARCH_FOR_SUPER_SIGNAL:graph_b1_s1',
                     #  'ENTRY_SIGNAL:graph_b1_s1',
                     #  'ENTRY_SIGNAL:graph_b1_s2',
                     #  'INIT_SIGNAL:graph_b1_s2',
                     #  'A:graph_b1_s2',
                     #  'A:graph_b1_s1',
                     #  'EXIT_SIGNAL:graph_b1_s2',
                     #  'SEARCH_FOR_SUPER_SIGNAL:graph_b1_s2',
                     #  'SEARCH_FOR_SUPER_SIGNAL:graph_b1_s2',
                     #  'ENTRY_SIGNAL:graph_b1_s2',
                     #  'INIT_SIGNAL:graph_b1_s2']

  # 7) If you need these transitions to happen very quickly, create a chart
  # using the HsmEventProcessor class instead of the InstrumentedHsmEventProcessor class
  chart = InstrumentedHsmEventProcessor()
  chart.start_at(graph_b1_s1)
  chart.dispatch(Event(signal=signals.A) # same transitions with no trace or spy
                                         # features

"""
import re
import pprint
import traceback  # try not use this if you can avoid it (it's fragile)
from copy         import copy
from functools    import wraps
from datetime     import datetime as stdlib_datetime
from contextlib   import contextmanager
from collections  import namedtuple, deque
from miros.event  import signals, return_status, Event


def pp(item):
  pprint.pprint(item)


SpyTuple = namedtuple('SpyTuple', ['signal',    'state',
                                   'hook',      'start',
                                   'internal',  'post_lifo',
                                   'post_fifo', 'post_defer',
                                   'recall',    'ignored',
                                   'datetime'])


def spy_tuple(signal     = None,
              state      = None,
              hook       = False,
              start      = False,
              internal   = False,
              post_lifo  = False,
              post_fifo  = False,
              post_defer = False,
              recall     = False,
              ignored    = False,
              datetime   = None,
              ):
  '''This is making it possible to have default settings in the SpyTuple,
     any attribute you need to over-write you can over-write by calling this
     function with it's name filled in with what you would like.'''
  if datetime is None:
    datetime = stdlib_datetime.now()
  return SpyTuple(signal=signal,
                  state=state,
                  start=start,
                  hook=hook,
                  internal=internal,
                  post_lifo=post_lifo,
                  post_fifo=post_fifo,
                  post_defer=post_defer,
                  recall=recall,
                  ignored=ignored,
                  datetime=datetime
                  )


def spy_on(fn):
  '''Instrument a state handling method'''
  @wraps(fn)
  def _spy_on(chart, *args):

    if len(args) == 1:
      e = args[0]
    else:
      e = args[-1]

    name = fn.__name__
    chart.state_name = name
    chart.state_fn   = fn
    # if the chart is not instrumented, don't try to wrap it
    if hasattr(chart, 'rtc') is False:
      # call the original handler and exit
      status = fn(chart, e)
      return status

    if(e.signal == signals.REFLECTION_SIGNAL):
      # We are no longer going to return a ReturnStatus object
      # instead we write the function name as a string
      status = name
      return status
    else:
      chart.rtc.spy.append("{}:{}".format(e.signal_name, name))

    # call the original handler
    status = fn(chart, e)

    if(signals.is_inner_signal(e.signal_name) is not True):
        # We have found a hook
        if(status is return_status.HANDLED):
          chart.rtc.spy.append("{}:{}:HOOK".format(e.signal_name, name))
          sr = spy_tuple(signal=e.signal_name, state=name, hook=True)
        else:
          if status is return_status.IGNORED:
            sr = spy_tuple(signal=e.signal_name, state=name, ignored=True)
          else:
            sr = spy_tuple(signal=e.signal_name, state=name)
    else:
      sr = spy_tuple(signal=e.signal_name, state=name, hook=True, internal=True)
    chart.rtc.tuples.append(sr)
    return status
  return _spy_on


def state_method_template(name):
  '''
  Used to create state chart methods with the register_signal_callback and
  register_parent API
  '''

  def base_state_method(chart, e):
    with chart.signal_callback(e, name) as fn:
      status = fn(chart, e)

    if(status == return_status.UNHANDLED):
      with chart.parent_callback(name) as parent:
        status, chart.temp.fun = return_status.SUPER, parent

    return status

  resulting_function = copy(base_state_method)
  resulting_function.__name__ = name
  resulting_function = spy_on(resulting_function)

  return resulting_function


# This is defined in the module name space so that inherited classes can access
# it
def spy_on_start(fn):
  '''instrument the on_start method into the spy log'''
  @wraps(fn)
  def _spy_on_start(self, initial_state):
    self.init_rtc()
    self.rtc.spy.append("START")
    sr = SpyTuple(
        signal="",       state="",
        start=True,      hook=False,
        internal=None,   post_lifo=True,
        post_fifo=False, post_defer=False,
        recall=False,    ignored=False,
        datetime=stdlib_datetime.now()
    )
    self.rtc.tuples.append(sr)
    # fn is start_at
    status = fn(self, initial_state)
    self.full.spy.extend(self.rtc.spy)
    return status
  return _spy_on_start


def append_fifo_to_spy(fn):
  @wraps(fn)
  def _append_fifo_to_spy(self, e):
    fn(self, e)
    if self.instrumented:
      self.rtc.spy.append("POST_FIFO:{}".format(e.signal_name))
  return _append_fifo_to_spy


# This is defined in the module name space so that inherited classes can access
# it
def trace_on_start(fn):
  '''instrument the on_start method into the trace log'''
  @wraps(fn)
  def _trace_on_start(self, initial_state):
    # fn is _spy_on_start
    status = fn(self, initial_state)
    if self.rtc.tuples[0].start:
      t = self.TraceTuple(
           datetime=stdlib_datetime.now(),
           start_state = 'top',
           signal      = None,
           payload     = None,
           end_state   =
             self.temp.fun(self,
               Event(signal=signals.REFLECTION_SIGNAL))
         )
      self.full.trace.append(t)
    return status
  return _trace_on_start


def append_queue_reflection_after_start(fn):
  @wraps(fn)
  def _append_queue_reflection_after_start(self, initial_state):
    result = fn(self, initial_state)
    self.rtc.spy.append(self.queue_reflection())
    self.full.spy.append(self.queue_reflection())
    return result
  return _append_queue_reflection_after_start


class Attribute():
  def __init__(self):
    pass


class HsmTopologyException(Exception):
  pass


class HsmEventProcessor():
  SPY_RING_BUFFER_SIZE = 500
  TRC_RING_BUFFER_SIZE = 150
  RTC_RING_BUFFER_SIZE = 250

  def __init__(self):
    # making the name space common
    '''set initial state of the '''
    # used by the event processor
    self.state = Attribute()
    self.temp  = Attribute()

  def start_at(self, initial_state):
    '''
    hsm = HsmEventProcessor()
    # build it
    hsm.start(starting_state_function)
    '''
    if self.state is None:
      self.state = Attribute()

    if self.temp is None:
      self.temp  = Attribute()

    self.state.fun = self.top
    self.temp.fun  = initial_state
    self.init()
    self.state_name = self.state.fun.__name__
    self.state_fn = self.state.fun

  def top(self, *args):
    '''top most state given to all HSM; treat it as an outside function'''
    status = return_status.IGNORED
    return status

  def init(self):
    '''triggers the top-most initial transition into a HSM

    This method is used at the beginning of an interaction with a statechart.
    It jumps into the initial state, then places all of the super state
    functions into a tpath variable.  Once this tpath is discovered, it calls the
    entry actions on each of the tpath items. Then it calls the init actions on
    the target, and if this init call requires a transition, it will perform all
    of the above actions again until the chart settles into its final initial
    state.

    This algorithm limits the design topology of your chart.  'init' signals
    that are going to be followed on the chart 'start_at' can only climb into
    the chart, they can not climb out.  If they do, this search routine will
    fail.

    '''
    topological_error = "impossible chart topology for HsmEventProcessor.init, "
    topological_error += "see HsmEventProcessor.init doc string for details"

    e = Event(signal=signals.SEARCH_FOR_SUPER_SIGNAL)
    tpath, outermost, max_index = [None], self.state.fun, 0

    # We will continue searching the chart until it stops requesting transitions
    while(True):  # outer while
      tpath[0], e, index = self.temp.fun, Event(signal=signals.SEARCH_FOR_SUPER_SIGNAL), 0
      previous_super = None
      while(self.temp.fun != outermost):
        index      += 1
        r           = self.temp.fun(self, e)
        if(previous_super == self.temp.fun):
          raise(HsmTopologyException(topological_error))
        if index > max_index:
          tpath.append(self.temp.fun)
          max_index = index
        else:
          tpath[index] = self.temp.fun
        previous_super = self.temp.fun

      self.temp.fun = tpath[0]
      # Now that we know what the tpaths are, starting for the outermost state,
      # enter each state until we reach our target state
      e = Event(signal=signals.ENTRY_SIGNAL)
      self.temp.fun = tpath[0]
      while(True):  # inner while
        # pre-decrement to remove outermost from our entry list
        index -= 1
        entery_fn = tpath[index]
        r = entery_fn(self, e)
        if(index <= 0):
          break  # inner while break condition

      # Now send it the init event, the init event could change our
      # self.temp.fun if it needs to transition. This means that we might have to
      # repeat the work done above.  Reset our outermost state to this state and
      # continue to delve into the chart
      outermost = tpath[0]
      e = Event(signal=signals.INIT_SIGNAL)
      r = outermost(self, e)

      if r != return_status.TRAN:
        break  # outer while break condition

    self.state.fun = outermost
    self.temp.fun  = outermost

  def dispatch(self, e):
    '''dispatches an event to a HSM.

    Processing an event represents one run-to-completion (RTC) step.  This code
    is largely based on the same processor which was written by Miro Samek in
    his book titled, "Practical Statecharts in C/C++: Event Driven Programming
    for Embedded Systems."  If you need to add features or functions add them as
    wrappers in inherited classes.  Try not to change this code too much since
    it is beautifully documentated within the sited book.

    Args:
      e (Event): The event to be dispatched to the hsm object

    Returns:
      None

    Example::
      chart = HsmEventProcessor()
      signals.append("A")
      chart.start_at(dispatch_graph_a1_s1)
      chart.dispatch(Event(signal=signals.A)

    Raises:
      HsmTopologyException: if a state function handler is malformed

    Useful mnemonics:

    S (source)           Which state is the source of the arrow in the diagram?
                         This variable is not actually defined, but it is
                         referenced in the comments with adjacent diagrams to keep
                         things clear.

                         Example::
                           S == dispatch_graph_f1_s0
                           # in the follow state function
                           def dispatch_graph_f1_s0(chart, e):
                             .
                             elif(e.signal == signals.C):
                               chart.trans(dispatch_graph_f1_s22)
                             .

    T (target)           What is the source aiming at?  Which state is pointed
                         to by the arrow in the diagram?  This variable is not
                         actually defined, but it is referenced in the comments.

                         This will be state handler that was an argument of the
                         trans call:

                         Example::
                           T == dispatch_graph_f1_s22
                           # in the follow state function
                           def dispatch_graph_f1_s0(chart, e):
                             .
                             elif(e.signal == signals.C):
                               chart.trans(dispatch_graph_f1_s22)
                             .

    S->super             The super state of S, the state in which S is wrapped
                         within.
    S->super->super..    The super..super state of S

    T->super             The super state of T, the state in which T is wrapped
                         within.
    T->super->super..    The super..super state of T

    self.state.fun       The current state before the dispatch occurred

    self.temp.fun        Before the search begins, this is T, but gets overwritten
                         during the search process by:
                           * any call to trans within state function will change
                             this.
                           * any call with a super signal will change this.

    lca                  least common ancestor.  The most outward state that S and
                         T have in common.  It is used to determine when we have
                         constructed the correct entry path to the target, called
                         the tpath.

    tpath                A list of state functions that are found during the
                         trans_ process.  If you call this list backwards,
                         starting from ip you will correctly enter toward T for
                         a given state chart topology.

    ip                   An index into the tpath.  All elements between 0 and ip
                         in the tpath are valid entry handlers which will be used
                         to approach T.

    iq                   Sometimes a bool indicating if the lac has been found,
                         sometimes a shadow of the ip index used to discover
                         which topology the statechart is configured in.  Only
                         used in more advanced topologies.

    topology             A graph characteristic that is shared well enough across
                         a set of graphs that part of this search algorithm can be
                         used to move from the S to the T.  There are 8 different
                         topologies, labeled topology_a..topology_h.  If you
                         would like to know more about them, reference the tests
                         directory where they are drawn or the trans_ method which
                         also has them described as diagrams in the comments.
    '''
    tpath       = [None, None, None]
    t, s, ip    = None, None, 0
    max_index   = 2

    entry_e, exit_e, super_e, init_e =                       \
              Event(signal=signals.ENTRY_SIGNAL),            \
              Event(signal=signals.EXIT_SIGNAL),             \
              Event(signal=signals.SEARCH_FOR_SUPER_SIGNAL), \
              Event(signal=signals.INIT_SIGNAL)

    # To begin with we assume that no action will occur and the T state is just
    # our current state.
    #
    # t contains T
    t = self.state.fun

    # Our contract
    assert(t is not None)
    assert(t == self.state.fun)

    # Determine if the chart can take action based on the event provided as
    # an argument to the method
    while(True):
      # if looped, place the super state into the s
      s = self.temp.fun
      # Event e was provided by the client, so we need to search outward in our
      # chart until we find a state that handles it, a state that handles it
      # will not return a return_state.SUPER status
      r = s(self, e)
      if r is None:
        raise(HsmTopologyException(
            "state handler {} is not returning a valid status".format(s)))
      if(r != return_status.SUPER):
        # S in now stored in s
        break

    # If we found a state that indicates that some action is required we
    # can now take action
    if(r >= return_status.TRAN):
      # tpath[0] contains T
      # tpath[1] contains t, self.state.fun (starting state)
      # tpath[2] contains S
      tpath[0], tpath[1], tpath[2] = self.temp.fun, t, s

      # Starting at our current state, move out to the S state, eventually
      # settling self.temp.fun on S, run all of the exit handlers we recurse
      # outward.
      #
      # t contains our current state at the start of the loop, then it is given
      # the self.state.fun->super.. Upon each iteration as it is passed the value
      # of self.temp.fun
      #
      # s contains S
      #
      # Note: This code and the topology_d code in trans_ allow provide the
      #       topology_h algorithm
      while(t != s):
        r = t(self, exit_e)
        if r is None:
          raise(HsmTopologyException(
              "state handler {} is not returning a valid status".format(t)))
        if(r == return_status.HANDLED):
          t(self, super_e)
        t = self.temp.fun

      # navigate all supported topologies
      # tpath will be over-written with entry values
      # ip will indicate at which point we should begin an entry path
      ip = self.trans_(tpath, max_index)

      # transition to history spy stuff placed here

      # If our trans_ method indicated that we need to enter into a state(s) we
      # do so now.  The tpath lower indexes contain inner states while the
      # higher indexes contain outer states.  So, we enter our outer states to
      # get toward the desired inner state.
      while(ip >= 0):
        tpath[ip](self, entry_e)
        ip -= 1

      # tpath[0] contains T
      # t is now set to T
      # self.temp.fun is set to T
      t, self.temp.fun = tpath[0], tpath[0]

      # We have entered T, now we need to work with its 'init' signal
      while(t(self, init_e) == return_status.TRAN):
        tpath[0], ip = self.temp.fun, 0
        self.temp.fun(self, super_e)

        while(self.temp.fun != t):
          ip += 1
          if ip > max_index:
            tpath.append(self.temp.fun)
            max_index = ip
          else:
            tpath[ip] = self.temp.fun
          self.temp.fun(self, super_e)

        self.temp.fun = tpath[0]

        while(True):
          tpath[ip](self, entry_e)
          ip -= 1
          if(ip <= 0):
            break
        t = tpath[0]

    elif(r is return_status.HANDLED):  # trans handled
      # This is the ultimate hook pattern
      pass
    else:
      pass
      # ignore the event since our chart doesn't handle it

    # t contains T
    self.state.fun = t
    self.temp.fun  = t

    self.state_name = t.__name__
    self.state_fn = t

  def trans(self, fn):
    '''sets a new function target and returns that transition required by engine'''
    self.temp.fun = fn
    return return_status.TRAN

  def trans_(self, tpath, max_index):
    '''execute a transition sequence in a hsm

    A helper function for the ```dispatch```.  It navigates through the possible
    supported topologies, navigating the chart just up to the point of entering
    the target hierarchy.  The target entry path is placed into the provided
    tpath list, and the depth of the entry path is provided as an output of the
    method.

    Args:
      tpath:      a list which will be populated with the entry path required
                  for the dispatch method to enter the T (target) state.

      max_index:  The maximum index used within the tpath list up until this
                  point.  If more space is required, an append is used to extend
                  the length of the tpath list, otherwise an element is assigned
                  to a given index location.

    Returns:
      ip:         An index into the tpath.  All elements between 0 and ip in the
                  tpath are valid entry handlers which will be used to approach T.

    If you don't understand what S/T are, read all of the mnemonics described in
    the ```dispatch``` docstring.

    To understand beyond this point you must first know what happens with
    the tpath, ip and iq. Consider this example::

     +-------- s1---------+
     | +-------s2-------+ |
     | | +-----s3-----+ | |
     | | | +---s4---+ | | |
     | | | | +-s5-+ | | | |
     | | | | T    | | | | | S
     | | | | +----+ | | | | top is the lca
     | | | +--------+ | | |
     | | +------------+ | |
     | +-+--------------+ |
     ++------------------++

    As trans_ searches, it will place state handlers into the tpath array
    These state handlers will be used to enter toward the target state once
    the lca has been found::

                  useful data <-+-> garbage data
                                |   collected in search
             +----+----+----+---+----+-----+-----+
      tpath: | s5 | s4 | s3 |s2 | s1 | top | s21 |
             +----+----+----+---+-/--+-----+-----+
                                  |
                                  +-- ip == 4

    The method that called trans_ already has a reference to the tpath so it
    doesn't need to be returned.  However, ip does need to be returned at it
    represents which state handlers will be entered.

    Returning the above from this method will tell ```dispatch``` to
    enter s1, enter s2, enter s3, enter s4, enter s5.

    iq is a bool, it represents if we have found the lca of S and T.  It is only
    used later in the method and it is not used outside of the method, so we
    only it when needed by the search.  It leave comments in the code describing
    its state, so you can understand what is going on.

    When the method begins t == T and s == S but these variable are then clobbered
    in the search and over-written with new meanings.  Their new meanings will
    be described in the comments where they are used, we will always draw our
    attention back to S and T and how they relate to a diagram.
    '''

    ip, iq = -1, 0  # no entry, no lca found
    # S is in tpath[2]
    # T is in tpath[0]
    t, s = tpath[0], tpath[2]

    exit_e, super_e =                                        \
              Event(signal=signals.EXIT_SIGNAL),             \
              Event(signal=signals.SEARCH_FOR_SUPER_SIGNAL)

    # +-S-T-+
    # |     +-+
    # |     | |
    # |     <-+
    # +-lca-+
    # (a) check source == target
    # pytest -m topology_a -s
    if(s is t):
      s(self, exit_e)  # exit the source
      ip = 0           # enter the target
      # iq = 1
    else:
      t(self, super_e)
      t = self.temp.fun
      # +--S-lca-+
      # | +-T--+ |
      # | |    <-+
      # | +----+ |
      # +--------+
      # t now contains T->super
      # s contains S
      # (b) check S == T->super
      # pytest -m topology_b -s
      if (s is t):
        ip = 0  # enter the target
        # iq = 1
      else:
        # find the super state of the source
        s(self, super_e)
        #  +-----lca-----+
        #  | +-S-+ +-T-+ |
        #  | |   | |   | |
        #  | |   +->   | |
        #  | |   | |   | |
        #  | +---+ +---+ |
        #  +-------------+
        # t now contains T->super
        # self.temp.fun contains S->super
        # (c) check S->super == T->super
        # pytest -m topology_c -s
        if(self.temp.fun == t):
          s(self, exit_e)
          ip = 0
          # iq = 1
        else:
          # +--T-lca-+
          # | +-S--+ |
          # | |    +->
          # | +----+ |
          # +--------+
          # self.temp.fun contains S->super
          # tpath[0] contains T
          # (d) check S->super == T
          # pytest -m topology_d -s
          if(self.temp.fun == tpath[0]):
            # leave ip as -1, that way no entry will occur
            s(self, exit_e)
            # iq = 1
          else:
            #  +--------S-lca-----+
            #  |+---------------+ |
            #  ||       .       | |
            #  ||   +-------+   | |
            #  ||   | +-T-+ |   | |
            #  || . | |   <-- . --+
            #  ||   | +---+ |   | |
            #  ||   +-------+   | |
            #  ||       .       | |
            #  |+---------------+ |
            #  +------------------+
            # (e) check S == T->super->super..
            # pytest -m topology_e -s
            iq, ip = 0, 1    # LCA not found yet, enter T and T->super
            tpath[1] = t     # tpath[1] contains T->super
            t = self.temp.fun  # t contains S->super
            r = tpath[1](self, super_e)

            if r is None:
              raise(HsmTopologyException(
                  "state handler {} is not returning a valid status".format(tpath[1])))

            while(r == return_status.SUPER):
              ip += 1
              # store our entry tpath
              if ip > max_index:
                tpath.append(self.temp.fun)
                max_index = ip
              else:
                tpath[ip] = self.temp.fun

              if(self.temp.fun is s):  # if we have found S
                iq =  1
                ip -= 1  # don't enter S since we are coming at it from the
                         # inside
                r = return_status.HANDLED  # terminate the loop
              else:
                r = self.temp.fun(self, super_e)

            if(iq is 0):
              # s contains S
              # set self.temp.fun to S->super
              s(self, exit_e)
              # If our previous search failed, the tpath is filled
              # with the entire state hierarchy.  So, we start on the
              # outside of the graph, and work our way in, checking if
              # S->super is the same as T->super->super..
              #
              # At this point of the search, the tpath has been filled
              # with the super states of our target and iq is a bool. If
              # it is 0 we must consider topology f:
              #    +---------------------------+
              #    |+-----------lca-----------+|
              #    ||      +--S--+            ||
              #    ||      |     +----------+ ||
              #    ||      +-----+          | ||
              #    ||+------------------+   | ||
              #    |||+---------------+ |   | ||
              #    ||||       .       | |   | ||
              #    ||||   +-------+   | |   | ||
              #    ||||   | +-T-+ |   | |   | ||
              #    |||| . | |   <-- . ------+ ||
              #    ||||   | +---+ |   | |     ||
              #    ||||   +-------+   | |     ||
              #    ||||       .       | |     ||
              #    |||+---------------+ |     ||
              #    ||+------------------+     ||
              #    |+-------------------------+|
              #    +---------------------------+
              # pytest -m topology_f -s
              # (f) check is S->super == T->super->super..
              # t contains S->super
              # tpath contains entry function handlers to get to T
              # tpath[ip] contains T->super->super..
              iq = ip
              r  = return_status.IGNORED
              while(True):
                if(t is tpath[iq]):  # is this the lca?
                  r = return_status.HANDLED
                  ip = iq - 1  # do not enter the lca
                  iq = -1
                else:
                  iq -= 1
                if(iq < 0):
                  break
            if(r != return_status.HANDLED):
              # +----------------------------+
              # |+----------lca-------------+|
              # || +-------------------+    ||
              # || | +---------------+ |    ||
              # || | |       .       | |    ||
              # || | |   +-------+   | |    ||
              # || | |   | +-S-+ |   | |    ||
              # || | | . | |   +-- . -----+ ||
              # || | |   | +---+ |   | |  | ||
              # || | |   +-------+   | |  | ||
              # || | |       .       | |  | ||
              # || | +---------------+ |  | ||
              # || +-------------------+  | ||
              # ||                        | ||
              # || +-------------------+  | ||
              # || | +---------------+ |  | ||
              # || | |       .       | |  | ||
              # || | |   +-------+   | |  | ||
              # || | |   | +-T-+ |   | |  | ||
              # || | | . | |   <-- . -----+ ||
              # || | |   | +---+ |   | |    ||
              # || | |   +-------+   | |    ||
              # || | |       .       | |    ||
              # || | +---------------+ |    ||
              # || +-------------------+    ||
              # |+--------------------------+|
              # +----------------------------+
              # (g) check each S->super->super.. == T->super->super
              # t contains S->super
              # tpath contains entry function handlers to get to T
              # ip initially will have the index equal to the maximum depth of
              # the chart relative to T, minus 1.
              r = return_status.IGNORED
              while(True):
                if(t(self, exit_e) == return_status.HANDLED):
                  t(self, super_e)
                # t contains S->super->super..
                t = self.temp.fun
                iq = ip
                while(True):
                  if(t == tpath[iq]):  # is this the lca?
                    r = return_status.HANDLED
                    ip = iq - 1  # do not enter the lca
                    iq = -1
                  else:
                    iq -= 1
                  if(iq < 0):
                    break
                if(r == return_status.HANDLED):
                  break
    return ip

  def is_in(self, fn_state_handler):
    '''tests if a hsm is in a given state'''
    result = False
    super_e = Event(signal=signals.SEARCH_FOR_SUPER_SIGNAL)
    while(True):
      if(self.temp.fun == fn_state_handler):
        result = True
        r = return_status.IGNORED
      else:
        r = self.temp.fun(self, super_e)
      if(r == return_status.IGNORED):
        break
    self.temp.fun = self.state.fun  # set our temp back to what it should be
    return result

  def child_state(self, fn_parent_state_handler):
    '''finds the child state of a given parent

    This method will only return a child state of a given handler, if the system
    is in a substate of the state being called::

      +---------- graph_e1_s1 -----------+
      | +-------- graph_e1_s2 -------+   |
      | | +------ graph_e1_s3 -----+ |   |
      | | | +---- graph_e1_s4 ---+ | |   |
      | | | |  +- graph_e1_s5 -+ | | |   |
      | | | |  |               | | | |   |
      | +-b->  |               <-----a---+
      | | | |  |               | | | |   |
      | | +c>  +---------------+ | | |   |
      +d> | +--------------------+ | |   |
      | | +------------------------+ |   |
      | +----------------------------+   |
      +----------------------------------+

      chart = HsmEventProcessor()
      chart.start_at(child_state_graph_e1_s5)
      chart.child_state(graph_e1_s5) #=> graph_e1_s5
      chart.child_state(graph_e1_s4) #=> graph_e1_s5
      chart.child_state(graph_e1_s3) #=> graph_e1_s4
      chart.dispatch(event=Event(signal=signals.D)

      # chart now in state graph_e1_s2
      chart.child_state(graph_e1_s5) #=> <CRASH!>
      chart.child_state(graph_e1_s2) #=> graph_e1_s2
      chart.child_state(graph_e1_s1) #=> graph_e1_s2 # which is wrong

    '''
    super_e   = Event(signal = signals.SEARCH_FOR_SUPER_SIGNAL)
    confirmed = False
    child     = self.state.fun
    self.temp.fun = self.state.fun
    while(True):
      if(self.temp.fun == fn_parent_state_handler):
        confirmed = True
        r = return_status.IGNORED
      else:
        child = self.temp.fun
        r = self.temp.fun(self, super_e)
      if(r == return_status.IGNORED):
        break
    self.temp.fun = self.state.fun
    assert(confirmed is True)
    return child

  def augment(self, **kwargs):
    """Used to add attributes to an hsm object

    Args:
      kwargs['other'](Mandatory other object): An another object for which you would
      like to add as an attribute of this object.

      kwargs['name'](Mandatory): The name that you would like to call this
      attribute, this argument must be a string

      kwargs['relationship'](Optional): Indicates if you want to also add this
      object as an attribute to the other class, using this object's name.  This
      option will only work if the other object also has an augment method that
      acts exactly the same as this one.

      ``Examples``
      alarm       = HsmEventProcessor(); alarm.name       = "alarm"
      time_keeper = HsmEventProcessor(); time_keeper.name = "time_keeper"
      alarm.augment(other=time_keeper, name="time_keeper")

      assert(alarm.time_keeper == time_keeper) # will be true

      inverter  = HsmEventProcessor(); inverter.name = "inverter"
      networker = HsmEventProcessor(); networker.name = "networker"
      inverter.augment(other=networker, name="net", relationship="mutual")

      assert(inverter.net == networker) # will be true
      assert(networker.inverter == inverter ) # will be true
    """

    relationship = None
    if("other" in kwargs):
      other = kwargs['other']
    if("name" in kwargs):
      name = kwargs['name']
    if("relationship" in kwargs):
      relationship = kwargs['relationship']

    if hasattr(self, name) is not True:
      setattr(self, name, other)
    else:
      pass
    if(relationship is not None and relationship is "mutual"):
      other.augment(other=self, name=self.name, relationship=None)


class InstrumentedHsmEventProcessor(HsmEventProcessor):
  '''

  '''
  SPY_RING_BUFFER_SIZE = 500
  TRC_RING_BUFFER_SIZE = 150
  RTC_RING_BUFFER_SIZE = 250

  def __init__(self):
    super().__init__()
    # used by spy/trace
    self.full  = Attribute()
    self.rtc   = Attribute()

    # used to build spy
    self.full.spy = deque(maxlen=HsmEventProcessor.SPY_RING_BUFFER_SIZE)

    # used to build trace
    self.full.trace = deque(maxlen=HsmEventProcessor.TRC_RING_BUFFER_SIZE)
    self.init_rtc()
    self.TraceTuple = namedtuple('TraceTuple',
                                            ['datetime',
                                             'start_state',
                                             'signal',
                                             'payload',
                                             'end_state'])

  def init_rtc(self):
    self.rtc.spy    = deque(maxlen=HsmEventProcessor.RTC_RING_BUFFER_SIZE)
    self.rtc.tuples = deque(maxlen=HsmEventProcessor.RTC_RING_BUFFER_SIZE)

  @trace_on_start
  @spy_on_start
  def start_at(self, initial_state):
    super().start_at(initial_state)

  def append_to_full_spy(fn):
    @wraps(fn)
    def _append_to_full_spy(self, e):
      self.rtc.spy.clear()
      # fn is dispatch
      fn(self, e)
      self.full.spy.extend(self.rtc.spy)
    return _append_to_full_spy

  def scribble(self, string):
    self.rtc.spy.append(string)

  def append_to_full_trace(fn):
    @wraps(fn)
    def is_signal_hooked_or_ignored(self):
      signal_name, hooked, ignored = "", False, True
      for sr in self.rtc.tuples:
        if sr.internal is False and sr.recall is False:
          signal_name = sr.signal
          dt = sr.datetime
          if sr.hook:
            hooked  = True
            ignored = False
            break
          ignored = sr.ignored
      return (signal_name, hooked, ignored, dt)

    def _append_to_full_trace(self, e):
      start_state = self.state.fun(self, Event(signal=signals.REFLECTION_SIGNAL))
      self.rtc.tuples.clear()
      # fn is append_to_full_spy
      fn(self, e)
      signal, hooked, ignored, dt = is_signal_hooked_or_ignored(self)
      if hooked is False and ignored is False:
        t = self.TraceTuple(
              datetime    = dt,
              start_state = start_state,
              signal      = signal,
              payload     = '',
              end_state   =
                self.state.fun(
                  self, Event(signal=signals.REFLECTION_SIGNAL)))

        self.full.trace.append(t)
    return _append_to_full_trace

  @append_to_full_spy
  @append_to_full_trace
  def dispatch(self, e):
    super().dispatch(e)


class HsmWithQueues(InstrumentedHsmEventProcessor):
  '''An Hsm that can post to itself and run to complete on each all of next_rtc.'''
  QUEUE_SIZE = 500

  def __init__(self, maxlen=QUEUE_SIZE, instrumented=True, priority=1):
    super().__init__()

    if instrumented:
      # see what is going on
      self.instrumented = True
    else:
      # run as fast as possible
      self.instrumented = False

    self.queue       = deque(maxlen = self.__class__.QUEUE_SIZE)
    self.defer_queue = deque(maxlen = self.__class__.QUEUE_SIZE)
    self.live_spy    = False
    self.live_trace  = False
    self.name        = None

    self.live_spy_callback   = self.__class__.live_spy_callback_default
    self.live_trace_callback = self.__class__.live_trace_callback_default

  @staticmethod
  def live_spy_callback_default(spy_line):
    print(spy_line)

  @staticmethod
  def live_trace_callback_default(trace_line):
    print(trace_line.replace("\n", ""))

  def register_live_spy_callback(self, live_spy_callback):
    self.live_spy_callback = live_spy_callback

  def register_live_trace_callback(self, live_trace_callback):
    self.live_trace_callback = live_trace_callback

  def print_spy_after_at_start_if_live(fn):
    @wraps(fn)
    def _print_spy_if_live(self, initial_state):
      # fn is _print_trace_if_live
      result = fn(self, initial_state)
      if self.instrumented and self.live_spy:
        for line in self.rtc.spy:
          self.live_spy_callback(line)
      return result
    return _print_spy_if_live

  def trace_tuple_to_formatted_string(self, tr):
    if self.name is None:
      name = 'None'
    else:
      name = self.name
    if tr.signal is None:
      signal = 'start_at'
    else:
      signal = tr.signal

    strace = "[{}] [{}] e->{}() {}->{}\n".format(
        stdlib_datetime.strftime(tr.datetime, "%Y-%m-%d %H:%M:%S.%f"),
        name,
        signal,
        tr.start_state,
        tr.end_state)
    return strace

  def print_trace_after_at_start_if_live(fn):
    @wraps(fn)
    def _print_trace_if_live(self, initial_state):
      # fn is next_rtc/start_at
      result = fn(self, initial_state)
      if self.instrumented and self.live_trace:
        strace = ""
        tr = self.full.trace[-1]
        strace  += self.trace_tuple_to_formatted_string(tr)
        self.live_trace_callback(strace)
      return result
    return _print_trace_if_live

  def print_spy_after_rtc_if_live(fn):
    @wraps(fn)
    def _print_spy_if_live(self):
      # fn is _print_trace_if_live
      result = fn(self)
      if self.instrumented and self.live_spy:
        for line in list(self.rtc.spy):
          self.live_spy_callback(line)
      return result
    return _print_spy_if_live

  def print_trace_after_rtc_if_live(fn):
    @wraps(fn)
    def _print_trace_if_live(self):
      # fn is next_rtc/start_at
      result = fn(self)
      if self.instrumented and self.live_trace:
        strace = "\n"
        tr = self.full.trace[-1]
        strace += self.trace_tuple_to_formatted_string(tr)
        self.live_trace_callback(strace)
      return result
    return _print_trace_if_live

  # print_spy_after_at_start_if_live calls print_trace_after_at_start_if_live
  # which calls append_queue_reflection_after_start which calls start_at.

  # append_queue_reflection_after_start does its work after start_at and
  # returns control to print_trace_after_at_start_if_live, which prints the
  # trace, then it returns control to print_spy_after_at_start_if_live which
  # prints the spy.
  @print_spy_after_at_start_if_live
  @print_trace_after_at_start_if_live
  @append_queue_reflection_after_start
  def start_at(self, initial_state):
    if self.instrumented:
      super().start_at(initial_state)
    else:
      HsmEventProcessor.start_at(self, initial_state)

  def current_state(self):
    if self.instrumented:
      cs = self.state.fun(self, Event(signals.REFLECTION_SIGNAL))
      return cs

  def dispatch(self, e):
    if self.instrumented:
      super().dispatch(e)
    else:
      HsmEventProcessor.dispatch(self, e)

  def queue_reflection(self):
    return "<- Queued:({}) Deferred:({})".format(len(self.queue), len(self.defer_queue))

  def append_lifo_to_spy(fn):
    @wraps(fn)
    def _append_lifo_to_spy(self, e):
      fn(self, e)
      if self.instrumented:
        self.rtc.spy.append("POST_LIFO:{}".format(e.signal_name))
    return _append_lifo_to_spy

  def append_defer_to_spy(fn):
    @wraps(fn)
    def _append_defer_to_spy(self, e):
      fn(self, e)
      if self.instrumented:
        self.rtc.spy.append("POST_DEFERRED:{}".format(e.signal_name))
    return _append_defer_to_spy

  def append_recall_to_spy(fn):
    @wraps(fn)
    def _append_recall_to_spy(self):
      if self.instrumented:
        if (len(self.defer_queue) != 0):
          sneak_peak = self.defer_queue[0]
          if sneak_peak is not None:
            sr = spy_tuple(signal=sneak_peak.signal_name, recall=True)
            self.rtc.spy.append("RECALL:{}".format(sneak_peak.signal_name))
            self.rtc.tuples.append(sr)
      e = fn(self)
      return e

    return _append_recall_to_spy

  def append_queue_reflection_to_spy(fn):
    @wraps(fn)
    def _append_queue_reflection_to_spy(self):
      # fn is _print_spy_if_live
      result = fn(self)
      if self.instrumented:
        self.rtc.spy.append(self.queue_reflection())
        self.full.spy.append(self.queue_reflection())
      return result
    return _append_queue_reflection_to_spy

  @append_fifo_to_spy
  def post_fifo(self, e):
    self.queue.append(e)

  @append_lifo_to_spy
  def post_lifo(self, e):
    self.queue.appendleft(e)

  @append_defer_to_spy
  def defer(self, e):
    self.defer_queue.append(e)

  @append_recall_to_spy
  def recall(self):
    e = None
    if(len(self.defer_queue) != 0):
      e = self.defer_queue.popleft()
      self.post_fifo(e)
    return e

  # print_spy_after_rtc_if_live calls print_trace_after_rtc_if_live which calls
  # append_queue_reflection_to_spy which calls next_rtc.

  # append_queue_reflection_to_spy does its work after next_rtc and returns
  # control to print_trace_after_rtc_if_live, which prints the trace, then it
  # returns control to print_spy_after_rtc_if_live which prints the spy.
  @print_spy_after_rtc_if_live
  @print_trace_after_rtc_if_live
  @append_queue_reflection_to_spy
  def next_rtc(self):
    self.rtc.spy.clear()
    action_taken = True
    if(len(self.queue) != 0):
      event = self.queue.popleft()
      self.dispatch(e=event)
    else:
      action_taken = False
    return action_taken

  def complete_circuit(self):
    action_taken = False
    while(len(self.queue) != 0):
      action_taken = self.next_rtc()
    return action_taken

  def clear_spy(self):
    if self.instrumented:
      self.full.spy.clear()

  def clear_trace(self):
    if self.instrumented:
      self.full.trace.clear()

  def trace(self):
    '''Output state transition information only:

    Example::
      print(chart.trace())
        [05:23:25.314420] [c] None: top->hsm_queues_graph_g1_s22
        [05:23:25.314420] [c] D: hsm_queues_graph_g1_s22->hsm_queues_graph_g1_s1
        [05:23:25.314420] [c] E: hsm_queues_graph_g1_s1->hsm_queues_graph_g1_s01
        [05:23:25.314420] [c] F: hsm_queues_graph_g1_s01->hsm_queues_graph_g1_s2111
        [05:23:25.314420] [c] A: hsm_queues_graph_g1_s2111->hsm_queues_graph_g1_s321
    '''
    strace = "\n"
    for tr in self.full.trace:
      strace += self.trace_tuple_to_formatted_string(tr)
    return strace

  def spy_rtc(self):
    return list(self.rtc.spy)

  def spy_full(self):
    return list(self.full.spy)

  def spy(self):
    return self.spy_full()

    return return_status.HANDLED

  def register_signal_callback(self, state_method, signal, fn):
    '''
    Example 1:
      def for_A(chart, e):
        print("Ah!")
        return return_status.HANDLED

      def state_function(chart, e):
        status = return_status.UNHANDLED
        with chart.lookup(e) as fn:
          result = fn(chart, e)
        else:
          status, chart.temp.fun = return_status.SUPER, chart.top
      return status

      ao.register_for_lookup(state_function, signals.A, for_A)

    Example 2:
      # if you are using the function template provided by this module

    '''
    def handled(chart, e):
      return return_status.HANDLED

    # ensure that if the user doesn't explicitly define the behavior for
    # 'entry', 'init' and 'exit' that these internal signals are at least
    # blocked and not let out of the function to fall out of the current
    # state method
    if hasattr(self, '_lookup') is False:
      self._lookup = {}
      self.register_signal_callback(state_method, signals.ENTRY_SIGNAL, handled)
      self.register_signal_callback(state_method, signals.INIT_SIGNAL, handled)
      self.register_signal_callback(state_method, signals.EXIT_SIGNAL, handled)

    if state_method.__name__ in self._lookup:
      self._lookup[state_method.__name__][signal] = fn
    else:
      self._lookup[state_method.__name__] = {}
      self._lookup[state_method.__name__][signal] = fn

  def register_parent(self, state_method, parent_method):
    if hasattr(self, '_parents') is False:
      self._parents = {}

    self._parents[state_method.__name__] = parent_method

  @contextmanager
  def parent_callback(self, state_method=None):
    if state_method is None:
      state_method = traceback.extract_stack(None, 3)[0][2]
    yield(self._parents[state_method])

  @contextmanager
  def signal_callback(self, e, name):
    '''
    with self.lookup(chart, e) as fn:
      result = fn(chart, e)
    '''
    if callable(name):
      key = name.__name__
    else:
      key = name

    def nothing_registered_for_signal(self, e):
      return return_status.UNHANDLED

    fn = nothing_registered_for_signal
    if(key in self._lookup):
      if(e.signal in self._lookup[key]):
        fn = self._lookup[key][e.signal]
    yield(fn)

  def to_code(self, state_method_name):
    '''
    Provides the equivalent flat code for items that have been written to a
    state_method written from a template.  This may be useful for debugging
    your code when you have used the 'register_signal_callback' and
    'register_parent' methods of this class.


    '''
    If_Blob = namedtuple('IfLadder', ['priority', 'signal_name', 'callback'])
    entry_priority, init_priority, other_priority, exit_priority = 1, 2, 3, 4

    def get_priority(signal_name):
      if signal_name == 'ENTRY_SIGNAL':
        priority = entry_priority
      elif signal_name == 'INIT_SIGNAL':
        priority = init_priority
      elif signal_name == 'EXIT_SIGNAL':
        priority = exit_priority
      else:
        priority = other_priority
      return priority

    def create_unordered_if_ladder(state_method):
      ifs = []
      for signal in self._lookup[state_method]:
        fn_name = self._lookup[state_method][signal].__name__
        signal_name = signals.name_for_signal(signal)
        priority = get_priority(signal_name)
        if_ladder = If_Blob(priority=priority, signal_name=signal_name, callback=fn_name)
        ifs.append(if_ladder)
      return ifs

    def create_ordered_if_ladder(unordered_ifs):
      ifs = sorted(unordered_ifs, key=lambda if_: if_.priority)
      return ifs

    def fill_missing_ifs(ordered_ifs):
      full_ordered_ifs = []
      entry  = [item for item in ordered_ifs if item.priority == entry_priority]
      init   = [item for item in ordered_ifs if item.priority == init_priority]
      others = [item for item in ordered_ifs if item.priority == other_priority]
      exit   = [item for item in ordered_ifs if item.priority == exit_priority]

      if len(entry) != 0:
        full_ordered_ifs.extend(entry)
      else:
        full_ordered_ifs.append(If_Blob(priority=entry_priority, signal_name='ENTRY_SIGNAL', callback=None))

      if len(init) != 0:
        full_ordered_ifs.extend(init)
      else:
        full_ordered_ifs.append(If_Blob(priority=init_priority, signal_name='INIT_SIGNAL', callback=None))

      if len(others) != 0:
        full_ordered_ifs.extend(others)

      if len(exit) != 0:
        full_ordered_ifs.extend(exit)
      else:
        full_ordered_ifs.append(If_Blob(priority=exit_priority, signal_name='EXIT_SIGNAL', callback=None))

      return full_ordered_ifs

    if callable(state_method_name):
      state_method = state_method_name.__name__
    else:
      state_method = state_method_name

    unordered_ifs = create_unordered_if_ladder(state_method)
    ordered_ifs = create_ordered_if_ladder(unordered_ifs)
    ifs = fill_missing_ifs(ordered_ifs)

    # to make it easier to test this code, we will start the string with a line
    # break
    code = "\n@spy_on\n"
    for i, if_tuple in enumerate(ifs):
      if i == 0:
        code += "def {}(chart, e):\n".format(state_method)
        code += "  status = return_status.UNHANDLED\n"
        if if_tuple.callback is None or if_tuple.callback is 'handled':
          code += "  if(e.signal == signals.{}):\n".format(if_tuple.signal_name)
          code += "    status = return_status.HANDLED\n"
        else:
          code += "  if(e.signal == signals.{}):\n".format(if_tuple.signal_name)
          code += "    status = {}(chart, e)\n".format(if_tuple.callback)
      else:
        if if_tuple.callback is None or if_tuple.callback is 'handled':
          code += "  elif(e.signal == signals.{}):\n".format(if_tuple.signal_name)
          code += "    status = return_status.HANDLED\n"
        else:
          code += "  elif(e.signal == signals.{}):\n".format(if_tuple.signal_name)
          code += "    status = {}(chart, e)\n".format(if_tuple.callback)

    parent_name = self._parents[state_method].__name__
    if parent_name == 'top':
      parent_name = 'chart.top'
    code += "  else:\n"
    code += "    status, chart.temp.fun = return_status.SUPER, {}\n".format(parent_name)
    code += "  return status\n"

    return code


@contextmanager
def stripped(log):
  '''Context manager used to compared trace/spy logs stripped of their timestamps

  The timestamp and chart name are expected to be on the front end of the
  trace/spy, between []:

    [2017-11-05 15:17:39.424492] [75c8c] e->BATTERY_CHARGE() armed->armed

  becomes:

    [75c8c] e->BATTERY_CHARGE() armed->armed

  We want to strip off the timestamp so that we can use the instrumented log as
  a specification for our design, this should become clear in the examples.

  Example 1:
    timestamp_trace1 = ao.trace()
    timestamp_trace2 = ao.trace()  # same to make a point

    with stripped(timestamped_trace1) as twt, \
         stripped(timestamped_trace2) as otw:

      for target_item, other_item in zip(twt, owt):
        assert(target_item == other_item)

  Example 2:
    with stripped('[2017-11-05 15:17:39.424492] [75c8c] e->BATTERY_CHARGE() armed->armed') /
      as swt:
      assert(swt == '[75c8c] e->BATTERY_CHARGE() armed->armed')

  '''
  def item_without_timestamp(item):
    '''
    [2017-11-05 15:17:39.424492] [75c8c] e->BATTERY_CHARGE() armed->armed
    -------- removed -----------
                                ------------ captured -------------------
    OR

       [2017-11-05 15:17:39.424492] [75c8c] e->BATTERY_CHARGE() armed->armed
    ===-------- removed -----------
          (front spaces removed)   ------------ captured -------------------

    '''
    m = re.match(r"[ ]{0,}\[[0-9-:. ]+\] (.+)$", item)
    if(m is not None):
      without_time_stamp = m.group(1)
    else:
      without_time_stamp = item
    return without_time_stamp

  targets = log.splitlines()
  if len(targets) > 1:
    stripped_target = []
    for target_item in targets:
      target_item = target_item.strip()
      if len(target_item) != 0:
        stripped_target_item = item_without_timestamp(target_item)
        stripped_target.append(stripped_target_item)
    yield(stripped_target)
  else:
    target = log
    yield(item_without_timestamp(target))





from collections import OrderedDict
from miros.singleton import SingletonDecorator
import json


# Not intended for export
class OrderedDictWithParams(OrderedDict):
  '''
  If your subclass <name_of_subclass> has the following init:
    def __init__(self,*args,**kwargs):

      self['RET_SUPER']     = 1
      self['RET_SUPER_SUB'] = 2
      self['UNHANDLED']     = 3

    Any object constructed from it will have to following attributes:
      obj = <name_of_subclass>
      obj.RET_SUPER     => 1
      obj.RET_SUPER_SUB => 2
      obj.UNHANDLED     => 3

    To post-pend an item to the object which will also have a named parameter:
      obj = <name_of_subclass>
      obj.append("NEW_NAMED_ATTRIBUTE")
      ob.NEW_NAMED_ATTRIBUTE => 4

  This of this class as being an extensible ENUM which isn't immutable.  All of
  the enums are wrapped up within an OrderedDict, so you get all of it's methods
  as well as the clean interface to the attribute.
  '''

  def append(self, string):
    if string in self:
      return
    else:
      self[string] = len(self) + 1

  def __getattr__(self, item):
    return self[str(item)]


# Not intended for export
class ReturnStatusSource(OrderedDictWithParams):

  '''
  A class which contains all of the state returns codes

  To construct the object
    state_returns = ReturnCodes()

  To get the number:
    state_returns.RET_SUPER => 1

  To add a return code:
    state_returns.append('RET_ZZ')
    state_returns.RET_ZZ => 12

  '''
  def __init__(self, *args, **kwargs):

    self['SUPER']     = 1
    self['SUPER_SUB'] = 2
    self['UNHANDLED'] = 3

    # handled and do not need to be bubbled up
    self['HANDLED']   = 4
    self['IGNORED']   = 5

    # entry/exit
    self['ENTRY']     = 6
    self['EXIT']      = 7

    # no side effects
    self['NULL']      = 8

    # transitions need to execute
    self['TRAN']      = 9
    self['TRAN_INIT'] = 10
    self['TRAN_HIST'] = 11
    self['TRAN_EP']   = 12
    self['TRAN_XP']   = 13


# Not intended for export
class SignalSource(OrderedDictWithParams):
  '''
  A class which contains all of the state signal types

  To get the basic system signals:
    signals = Signal()

  To append a new signal
    signals = Signal.append('OVEN_OFF')

  To get the number:
    signal.ENTRY_SIGNAL => 1
    signal['OVEN_OFF']  => 12
    signal.OVER_OFF     => 12

  '''
  def __init__(self, *args, **kwargs):

    self['ENTRY_SIGNAL']            = 1
    self['EXIT_SIGNAL']             = 2
    self['INIT_SIGNAL']             = 3
    self['REFLECTION_SIGNAL']       = 4
    self['SEARCH_FOR_SUPER_SIGNAL'] = 5

  def append(self, string):
    if string in self:
      return
    else:
      self[string] = len(self) + 1

  def is_inner_signal(self, other):
    def is_number_an_internal_signal(number):
      result = False
      if number in list(self.values())[0:self.SEARCH_FOR_SUPER_SIGNAL]:
        result = True
      return result

    result = False
    if(isinstance(other, str)):
      try:
        other = self[other]
        result = is_number_an_internal_signal(other)
      except:
        pass
    elif(isinstance(other, int)):
      try:
        result = is_number_an_internal_signal(other)
      except:
        pass
    return result

  def name_for_signal(self, signal):
    '''
    get the name of a signal number as a string
    '''
    signal_name = list(self.keys())[list(self.values()).index(signal)]
    return signal_name

  def __getattr__(self, item):
    value = None
    try:
      value = self[str(item)]
    except:
      self.append(item)
      value = self[str(item)]

    return value


'''
Defining the signals used by this package and all of the packages that
reference it.  Think of this as a singleton or a growing enumeration.
'''
# Signal is a singleton
Signal = SingletonDecorator(SignalSource)
signals_exist = 'signals' in locals()
if signals_exist is False:
  signals = Signal()

# ReturnStatus is a singleton
ReturnStatus = SingletonDecorator(ReturnStatusSource)
status_exist = 'return_status' in locals()
if status_exist is False:
  return_status = ReturnStatus()


class Event(OrderedDictWithParams):
  '''
  An event should be constructed, used, then garbage collected.  An event is a
  temporary thing.  However if an event uses a signal that hasn't been seen
  before, that signal will be added to the list of global signals as a new
  enumerated value.

  # Make an event (this should happen internally):
    e = Event(signal=signals.ENTRY_SIGNAL) # existing signal
    assert(e.signal == signals.ENTRY_SIGNAL)
    assert(e.signal_name == 'ENTRY_SIGNAL')

  # Make an event, which will construct a signal internally:
    e = Event(signal ='OVEN_OFF', payload='any object can go here') # new signal
    assert(e.signal == 5) # if it is the first unseen signal in the system
    assert(e.signal_name == 'OVEN_OFF')
    assert(signals.OVER_OFF == 5)

  '''
  def __init__(self, signal, payload=None):
    global signals

    self.payload = payload

    if signal in signals.values():
      self.signal = signal
      for key, value in signals.items():
        if value == signal:
          self.signal_name = key
          break

    elif isinstance(signal, str):
      signals.append(signal)
      self.signal_name = signal
      # over-write the signal string as the signal name
      self.signal      = signals[signal]
    else:
      raise("signal must be of type string or Signal")

  def has_payload(self):
    '''
    Example:
      event = Event(signal=signals.Mary, payload=[1,2,3])
      event.has_payload() # => True
    '''
    result = True
    if self.payload is None:
      result = False
    return result

  def __str__(self):
    '''
    Provides a string representation of an event object.

    Example:
      event = Event(signal=signals.Mary, payload=[1,2,3])
      pdb> print(event) #=> Mary::<[1, 2, 3]>
    '''
    if self.payload is None:
      result = "{}::<>".format(self.signal_name)
    else:
      result = "{}::<{}>".format(self.signal_name, str(self.payload))
    return result

  def __repr__(self):
    '''
    Provides an internal representation of an event object.

    Example:
      event = Event(signal=signals.Mary, payload=[1,2,3])
      pdb> event #=> Mary:8::<[1, 2, 3]>

    Note: The number after the signal name is the internal number used for
    tracking the signal.

    '''
    if self.payload is None:
      result = "{}:{}::<>".format(self.signal_name, self.signal)
    else:
      result = "{}:{}::<{}>".format(self.signal_name, self.signal, str(self.payload))
    return result

  @staticmethod
  def dumps(event):
    '''
    Required for serialization prior to sending an event across an IO
    stream/networks.

    Due to the dynamic nature of the Event/signals object, an event object can
    not just be pickled.  Instead the object must be passed through this 'Event.dumps'
    function, then de-serialized using the 'Event.loads' function.

    The numbering parity of the signal is not guaranteed across a network, but,
    the signal name will be unique within all of the signals objects expressed on each
    machine.  This is because signals are constructed dynamically.  So if one
    machine defines the signal 'A' before 'B' and another just defines 'B'.  The
    internal signals numbers for 'B' will not be the same for each machine.
    This doesn't matter, since it is the signal name 'B' that distinguishes the
    signal within a statechart and each machine will have the 'B' signal defined
    once.

    This 'dumps' method pulls out the signal name and payload, creates a
    dictionary from them then turns it into a json object.   The signal number
    is ignored during the process.

    Example:
      json_event = Event.dumps(Event(signal=signals.Mary))

    '''
    signal_name = event.signal_name
    payload = None
    if event.payload is not None:
      payload = event.payload

    event_as_dict = {'signal_name' : signal_name,
                     'payload' : payload}
    return json.dumps(event_as_dict)

  @staticmethod
  def loads(json_event):
    '''
    De-serializes a serialized Event object.  An event object can not be
    serialized using pickle due to it's dynamic nature.  For this reason the
    'Event.dumps' method provides custom serialization for events.  This 'loads'
    method is for deserializing a serialized event object.  See the example for
    clarity.

    Example:
      json_event = Event.dumps(Event(signal=signals.Mary))
      event = Event.loads(json_event)
      print(event) # => Mary::<>

    '''
    event_as_dict = json.loads(json_event)
    signal_name = event_as_dict['signal_name']
    payload     = event_as_dict['payload']

    event = Event(signal = signal_name, payload=payload)
    return event
